"""
crawler_service.py: Utility functions used by the crawler service.
"""
__author__ = "Jason Wolosonovich <jason@avaland.io>"
__license__ = "BSD 3 clause"
from urllib.parse import urlparse
import re
import json

import requests
import google.cloud.language as language
from bs4 import BeautifulSoup

from .classes import MetadataMixin


HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                  "AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/61.0.3163.100 Safari/537.36"
}


def is_email_link(href=None):
    """Utility function to determine whether the supplied href attribute
    is an email link."""
    print("email_link()")
    return href and 'mailto:' in href


def is_tel_link(href=None):
    """Utility function to determine whether the supplied href
    attribute is a telephone link."""
    print("is_tel_link()")
    return href and 'tel:' in href


def is_social_link(href=None):
    """Utility function to determine whether the supplied href
    attribute is a social media link."""
    print("is_social_link()")
    return href and (
            'facebook' in href or 'twitter' in href or 'pinterest'
            in href or 'linkedin' in href)


def get_emails(document=None):
    """Utility function to return emails from a document."""
    print("get_emails()")
    _r = r"'<\S+?>'"
    # this caught false-positives like 'rd@context'
    # _r = r"[\w\.-]+@[\w\.-]+"
    emails = list(
        set(
            re.findall(
                _r,
                document.text) +
            [
                a["href"] for a in
                document.find_all(
                    "a",
                    href=is_email_link
            )
             ]
        )
    )
    return emails


def get_phones(document=None):
    """Utility function that returns any phone numbers found within the
    supplied document."""
    print("get_phones()")
    _r = r'(?:(?:\+?([1-9]|[0-9][0-9]|[0-9][0-9][0-9])' \
         r'\s*(?:[.-]\s*)?)?(?:\(\s*([2-9]1[02-9]|[2-9][02-8]1|[2-9]' \
         r'[02-8][02-9])\s*\)|([0-9][1-9]|[0-9]1[02-9]|[2-9][02-8]1|' \
         r'[2-9][02-8][02-9]))\s*(?:[.-]\s*)?)?([2-9]1[02-9]|[2-9]' \
         r'[02-9]1|[2-9][02-9]{2})\s*(?:[.-]\s*)?([0-9]{4})(?:\s*' \
         r'(?:#|x\.?|ext\.?|extension)\s*(\d+))?'
    regex_captured_numbers = re.findall(_r, document.text)
    phones = list(set([''.join(match) for
                       match in regex_captured_numbers] +
                      [a["href"] for
                       a in document.
                      find_all("a", href=is_tel_link)]))
    # TODO: hack; fix in regex
    phones = [phone.strip("tel:") for phone in phones]
    return phones


def get_socials(document=None):
    """Utility function that returns social media links from the
    supplied document."""
    print("get_socials()")
    # TODO: add regex back in later
    socials = list(set(
        # re.findall(r'[\w\.-]+@[\w\.-]+', document.text) + [
        [a["href"] for a in
         document.find_all("a", href=is_social_link)]))

    return socials

# TODO: needs implemented
def get_url(ip_address=None):
    """Converts the supplied IP address to the associated URL for
    crawling and parsing."""
    print("get_url()")
    # TODO: figure out how to overcome CDNs (like cloudflare) which
    # obscure the DNS info though perhaps that is a sign that the
    # company is probably too big for our product


def get_content_class(content=None):
    """Utility function that takes website content and returns the
    classification of that text by Google."""
    try:
        print("get_content_class()")
        language_client = language.LanguageServiceClient()

        document = language.types.Document(
            content=content,
            type=language.enums.Document.Type.PLAIN_TEXT)

        response = language_client.classify_text(document)

        categories = response.categories

        result = {}

        for category in categories:
            result[category.name] = category.confidence

        # google may not return anything
        if len(result) == 0:
            result["No Category Available"] = 0.0
    except Exception as e:
        print(e)
        result = {
            "An Exception Occurred": 0.0
        }
    return result


def get_wp_plugins(document=None):
    """Retrieves a list of the installed WP plugins."""
    print("get_wp_plugins()")
    r = r"""(?<=wp-content\/plugins\/)(.*?)(?=\/)"""
    plugin_links = document \
        .find_all('link',
                  {'href': re.compile(r)})
    plugins_list = [re.findall(r, tag["href"])[0] for tag in
                    plugin_links]
    plugins_list = list(set(plugins_list))
    return plugins_list


def get_wp_themes(document=None):
    """Returns the WP theme used on the site."""
    print("get_wp_themes()")
    r = r"""(?<=wp-content\/themes\/)(.*?)(?=\/)"""
    themes = document.find_all("link",
                               {"href": re.compile(r)})
    themes_list = [re.findall(r, tag["href"])[0] for tag in themes]
    themes_list = list(set(themes_list))
    return themes_list


def get_squarespace_context(document=None):
    """Returns the Squarespace template used on the site."""
    # this content contains all kinds of useful firmographic data.
    try:
        script = document.find(
            "script",
            {"data-name": "static-context"}
        )\
        .getText(strip=True)

        # now parse out the context object
        context = re.search(
            "(?ms)Static\.SQUARESPACE_CONTEXT = ({.+?});",
            script,
        ).group(1)
        context = json.loads(context)
    except json.decoder.JSONDecodeError as e:
        print(e)
        context = None
    except Exception as e:
        print(e)
        context = None
    return context


def get_all_links(document=None):
    """Function that returns all the links found on the visited page."""
    print("get_all_links()")
    all_links = [a["href"]
                 for a in document.find_all("a", href=True)]
    # remove empty strings
    all_links = [link for link in all_links if len(link) > 0 and
                 link != '/']
    all_links = list(set(all_links))
    return all_links


def get_internal_links(url=None, links=None):
    """Function that returns the internal links from the website."""
    print("get_internal_links()")
    internal_links_ = list()
    for link in links:
        if link is None:
            continue
        print(link)
        url_host = urlparse(url).hostname
        domain = urlparse(url).hostname.split(".")[1]
        print(domain)
        link_host = urlparse(link).hostname
        print(link_host)
        if link.startswith("/"):
            print("True")
            internal_links_.append(link)
        if url_host == link_host:
            internal_links_.append(link)
        if link_host is not None and domain in link_host:
            internal_links_.append(link)
        else:
            continue

    return internal_links_


def get_external_links(internal_links=None,  all_links=None):
    """Function that returns the external links from a site."""
    print("get_external_links()")
    # TODO: clean this up to remove incorrect links like jquery,
    # tel:, etc.
    _r = r"(http|ftp|https)://([\w_-]+(?:(?:\.[\w_-]+)+))" \
         r"([\w.,@?^=%&:/~+#-]*[\w@?^=%&/~+#-])?"
    links = [link for link in all_links
             if link not in internal_links]
    # do some additional cleanup to remove jquery anchors, etc.
    external_links = [re.findall(_r, link) for link in links]
    # remove empty lists
    external_links = [link for link in external_links
                      if len(link) > 0]
    # reconstruct original links from remaining tuples
    external_links = [l[0][0] + "://" + l[0][1] + l[0][2]
                      for l in external_links]
    return external_links


def get_hrefs(document=None):
    """Function to return the email, phone and social data from a 
    website."""
    print("get_hrefs()")
    # attempt to fill in the attributes
    emails_ = get_emails(document=document)
    phones_ = get_phones(document=document)
    socials_ = get_socials(document=document)
    return emails_, phones_, socials_


def get_keywords_and_description(document=None):
    """Extracts the keywords and description from the `meta`
    tags."""
    print("get_keywords_and_description()")
    keywords_ = document \
        .find_all(name="meta",
                  attrs={"name": "keywords"})

    if len(keywords_) != 0:
        keywords_ = keywords_[0]["content"] \
            .split(",")

        keywords_ = [keyword.lstrip().rstrip() for keyword in keywords_]

    description_ = document \
        .find_all(name="meta",
                  attrs={"name": "description"})

    if len(description_) != 0:
        description_ = description_[0]["content"].replace("\r", " ")
    return keywords_, description_


def crawl_and_parse(url=None):
    """Utility function to crawl and return the HTML from the
    supplied domain."""
    print("crawl_and_parse()")
    try:
        resp = requests.get(url=url,
                            headers=HEADERS)
        resp.raise_for_status()
    except requests.exceptions.HTTPError as e:
        resp = "Error: {}".format(e)
    except Exception as e:
        resp = "Error: {}".format(e)

    # OK
    if resp.status_code == 200:
        return BeautifulSoup(resp.content,
                             "html.parser")
    else:
        return "error: {}".format(resp)


def make_crawler_gcs_payload(request=None):
    """Utility function to flatten and parse the fields we need in
    BQ."""
    print("make_crawler_gcs_payload()")
    p = MetadataMixin()
    r = request.copy()

    p["domain"] = r["domain"]
    p["url"] = r["url"]
    p["html_string"] = r["document"]
    p["all_links"] = r["all_links"]
    p["internal_links"] = r["internal_links"]
    p["external_links"] = r["external_links"]
    p["href_emails"] = r["href_emails"]
    p["href_phones"] = r["href_phones"]
    p["href_socials"] = r["href_socials"]
    p["meta_keywords"] = r["meta_keywords"]
    p["ip_revealed"] = r["ip_revealed"]
    p["fuzzy_match"] = r["fuzzy_match"]
    if len(r["meta_description"]) != 0:
        p["meta_description"] = r["meta_description"].replace("\r", "")
    else:
        p["meta_description"] = None
    # add the metadata we injected to the request along the way
    p["refinery_company_id"] = r["refinery_company_id"]
    p["refinery_person_id"] = r["refinery_person_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["sfdc_lead_id"] = r["sfdc_lead_id"]
    p["sfdc_contact_id"] = r["sfdc_contact_id"]
    p["sfdc_asset_id"] = r["sfdc_asset_id"]
    p["sfdc_acct_id"] = request["sfdc_acct_id"]
    p["sfdc_oppty_id"] = request["sfdc_oppty_id"]
    p["app_name"] = request["app_name"]
    # TODO: should we keep all the content classifications if > 1 are
    # returned?
    if "No Category Available" in r["content_classification"].keys():
        p["classification_category"] = "No Category Available"
        p["classification_confidence"] = 0.0
        p["tier1_classification"] = "No Category Available"
        p["tier2_classification"] = None
        p["tier3_classification"] = None
    elif "An Exception Occurred" in r["content_classification"].keys():
        p["classification_category"] = "An Exception Occurred"
        p["classification_confidence"] = 0.0
        p["tier1_classification"] = "An Exception Occurred"
        p["tier2_classification"] = None
        p["tier3_classification"] = None
    else:
        max_class = None
        max_likelihood = 0.0
        for cat, prob in r["content_classification"].items():
            if prob > max_likelihood:
                max_likelihood = prob
                max_class = cat

            p["classification_category"] = max_class
            p["classification_confidence"] = max_likelihood
            # there are 1, 2 or 3 possible levels deep for the
            # classification
            # https://cloud.google.com/natural-language/docs/categories
            cat_list = max_class.split("/")[1:]
            if len(cat_list) == 1:
                p["tier1_classification"], \
                p["tier2_classification"], \
                p["tier3_classification"] = cat_list[0], None, None
            if len(cat_list) == 2:
                (
                    p["tier1_classification"],
                    p["tier2_classification"]
                ), \
                p["tier3_classification"] = cat_list, None
            if len(cat_list) == 3:
                p["tier1_classification"], \
                p["tier2_classification"], \
                p["tier3_classification"] = cat_list
    print("payload: {}".format(p))
    return p


def make_crawler_bq_payload(request=None):
    """Utility function to flatten and parse the fields we need in
    BQ."""
    print("make_crawler_gcs_payload()")
    p = MetadataMixin()
    r = request.copy()

    p["domain"] = r["domain"]
    p["url"] = r["url"]
    p["html_string"] = r["html_string"]
    p["ip_revealed"] = r["ip_revealed"]
    p["fuzzy_match"] = r["fuzzy_match"]
    p["all_links"] = " >>> ".join(
        r["all_links"]
    )
    p["internal_links"] = " >>> ".join(
        r["internal_links"]
    )
    p["external_links"] = " >>> ".join(
        r["external_links"]
    )
    p["href_emails"] = " >>> ".join(
        r["href_emails"]
    )
    p["href_phones"] = " >>> ".join(
        r["href_phones"]
    )
    p["href_socials"] = " >>> ".join(
        r["href_socials"]
    )
    p["meta_keywords"] = " >>> ".join(
        r["meta_keywords"]
    )
    p["meta_description"] = r["meta_description"]
    # add the metadata we injected to the request along the way
    p["refinery_company_id"] = r["refinery_company_id"]
    p["refinery_person_id"] = r["refinery_person_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["sfdc_lead_id"] = r["sfdc_lead_id"]
    p["sfdc_contact_id"] = r["sfdc_contact_id"]
    p["sfdc_asset_id"] = r["sfdc_asset_id"]
    p["sfdc_oppty_id"] = r["sfdc_oppty_id"]
    p["sfdc_acct_id"] = r["sfdc_acct_id"]
    p["app_name"] = r["app_name"]
    p["classification_confidence"] = r["classification_confidence"]
    p["tier1_classification"] = \
        r["tier1_classification"]
    p["tier2_classification"] = \
        r["tier2_classification"]
    p["tier3_classification"] = \
        r["tier3_classification"]
    print("payload: {}".format(p))
    return p


def wordpress_routines(request=None, has_html=True, is_parsed=False):
    """Utility function containing all the wordpress enrichment 
    routines."""
    print("wordpress_routines()")

    # container for holding the crawler results
    resp = MetadataMixin()

    # TODO: clean up the logic here
    if has_html is False:
        resp["url"] = request["url"]
        # crawl the site and return the html for parsing
        resp["document"] = crawl_and_parse(url=resp["url"])
    elif has_html is True:
        # add the domain and url to the metadata container
        # resp["domain"] = request["domain"]
        if is_parsed is True:
            resp["document"] = request["document"]
        elif is_parsed is False:
            resp["document"] = BeautifulSoup(request["document"],
                                             "html.parser")

    try:
        # get the word press themes
        resp["wp_themes"] = get_wp_themes(document=resp["document"])
    except Exception as e:
        print(e)
        resp["wp_themes"] = "ERROR"

    try:
        # get the word press plugins
        resp["wp_plugins"] = get_wp_plugins(document=resp["document"])
    except Exception as e:
        print(e)
        resp["wp_plugins"] = "ERROR"
    return resp


def squarespace_routines(request=None):
    """Utility function containing"""
    print("squarespace_routines()")

    # container for holding the crawler results
    resp = MetadataMixin()

    # add the domain and url to the metadata container
    resp["domain"] = request["domain"]
    resp["url"] = request["url"]

    # crawl the site and return the html for parsing
    resp["document"] = crawl_and_parse(url=resp["url"])

    # if the site is squarespace, we'll get the context object and
    # parse the info in it
    resp["squarespace_context"] = get_squarespace_context(
        document=resp["document"])

    return resp


def domain_routines(request=None):
    """Utility function containing the normal enrichment routines for
    the webcrawler."""
    print("domain_routines()")

    # container for holding the crawler results
    resp = MetadataMixin()

    # add the domain and url to the metadata container
    resp["domain"] = request["domain"]
    resp["url"] = request["url"]

    # crawl the site and return the html for parsing
    resp["document"] = crawl_and_parse(url=resp["url"])

    # get all links
    resp["all_links"] = get_all_links(document=resp["document"])

    # get the internal links
    resp["internal_links"] = \
        get_internal_links(url=resp["url"],
                               links=resp["all_links"])

    # get the external links
    resp["external_links"] = \
        get_external_links(internal_links=resp["internal_links"],
                               all_links=resp["all_links"])

    # get the phone, email and social data
    resp["href_emails"], resp["href_phones"], resp["href_socials"] = \
        get_hrefs(document=resp["document"])

    # get the keyword and description
    resp["meta_keywords"], resp["meta_description"] = \
        get_keywords_and_description(document=resp["document"])

    # get the content classification
    resp["content_classification"] = \
        get_content_class(content=resp["document"].text)

    # convert bs4 soup to plain string so we can convert it to json
    resp["document"] = resp["document"].prettify()

    return resp


# TODO: can we abstract this to fit all cases (tags, tech, wp)
def make_wordpress_asset_history_bq_payload(request=None):
    """Utility function that creates a payload for the Wordpress
    bucket."""
    print("make_wordpress_asset_history_payload()")
    r = request.copy()
    p = MetadataMixin()

    p["refinery_company_id"] = r["refinery_company_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["html_used"] = r["document"]
    # TODO: add additional lists here (wix, squarespace, etc.)
    if len(r["wp_themes"]) != 0:
        p["wp_themes"] = r["wp_themes"]
    else:
        p["wp_themes"] = list()
    if len(r["wp_plugins"]) != 0:
        p["wp_plugins"] = r["wp_plugins"]
    print("payload: {}".format(p))
    return p


def make_crawler_squarespace_payload(request=None):
    """Extracts useful fields from the Static.SQUARESPACE_CONTEXT
    object."""
    print("make_crawler_squarespace_payload()")
    r = request["squarespace_context"].copy()
    p = MetadataMixin()
    p["refinery_company_id"] = request["refinery_company_id"]
    p["refined_at"] = request["refined_at"]
    p["refined_date"] = request["refined_date"]
    p["domain"] = request["domain"]
    p["url"] = request["url"]
    p["ip_revealed"] = request["ip_revealed"]
    p["fuzzy_match"] = request["fuzzy_match"]

    # there is no guarantee any of these will be available in the
    # payload
    try:
        p["template_id"] = r["templateId"]
    except KeyError as e:
        p["template_id"] = None
    try:
        p["authentic_url"] = r["website"]["authenticUrl"]
    except KeyError as e:
        p["authentic_url"] = None
    try:
        p["base_url"] = r["website"]["baseUrl"]
    except KeyError as e:
        p["base_url"] = None
    try:
        p["content_modified_at"] = r["website"]["contentModifiedOn"]
    except KeyError as e:
        p["content_modified_at"] = None
    try:
        p["internal_url"] = r["website"]["internalUrl"]
    except KeyError as e:
        p["internal_url"] = None
    try:
        p["language"] = r["website"]["language"]
    except KeyError as e:
        p["language"] = None
    try:
        p["time_zone"] = r["website"]["timeZone"]
    except KeyError as e:
        p["time_zone"] = None
    try:
        p["time_zone_abbr"] = r["website"]["timeZoneAbbr"]
    except KeyError as e:
        p["time_zone_abbr"] = None
    try:
        p["utc_offset"] = r["website"]["TimeZoneOffset"] / 3.6e+6
    except KeyError as e:
        p["utc_offset"] = None
    try:
        p["site_description"] = r["website"]["siteDescription"].replace(
            "\r", "").replace("\n", "").replace("<p>", "")
    except KeyError as e:
        p["site_description"] = None
    try:
        p["site_title"] = r["website"]["siteTitle"]
    except KeyError as e:
        p["site_title"] = None
    try:
        p["primary_domain"] = r["website"]["primaryDomain"]
    except KeyError as e:
        p["primary_domain"] = None
    try:
        # TODO: this is a list so iterate through when adding to BQ
        p["social_accounts"] = r["website"]["socialAccounts"]
    except KeyError as e:
        p["social_accounts"] = None
    try:
        p["state"] = r["websiteSettings"]["state"]
    except KeyError as e:
        p["state"] = None
    try:
        p["supported_currencies"] = \
        r["websiteSettings"]["storeSettings"] \
            ["currenciesSupported"]
    except KeyError as e:
        p["supported_currencies"] = None
    try:
        p["default_currency"] = r["websiteSettings"]["storeSettings"] \
            ["defaultCurrency"]
    except KeyError as e:
        p["default_currency"] = None
    try:
        p["express_checkout_enabled"] = \
        r["websiteSettings"]["storeSettings"] \
            ["expressCheckout"]
    except KeyError as e:
        p["express_checkout_enabled"] = None
    try:
        p["light_cart_enabled"] = r["websiteSettings"]["storeSettings"] \
            ["useLightCart"]
    except KeyError as e:
        p["light_cart_enabled"] = None
    try:
        p["squarespace_website_id"] = \
        r["websiteSettings"]["storeSettings"] \
            ["websiteId"]
    except KeyError as e:
        p["squarespace_website_id"] = None
    try:
        p["mailing_list_opt_in_enabled"] = \
        r["websiteSettings"]["storeSettings"] \
            ["enableMailingListOptInByDefault"]
    except KeyError as e:
        p["mailing_list_opt_in_enabled"] = None
    try:
        p["business_name"] = \
            r["websiteSettings"]["storeSettings"] \
                ["businessName"]
    except KeyError as e:
        p["business_name"] = None
    try:
        p["business_address_line_1"] = \
            r["websiteSettings"]["storeSettings"] \
                ["contactLocation"]["addressLine1"]
    except KeyError as e:
        p["business_address_line_1"] = None
    try:
        p["business_address_line_2"] = \
            r["websiteSettings"]["storeSettings"] \
                ["contactLocation"]["addressLine2"]
    except KeyError as e:
        p["business_address_line_2"] = None
    try:
        p["business_address_country"] = \
            r["websiteSettings"]["storeSettings"] \
                ["contactLocation"]["addressCountry"]
    except KeyError as e:
        p["business_address_country"] = None
    try:
        p["typekit_id"] = r["website"]["typekitId"]
    except KeyError as e:
        p["typekit_id"] = None
    try:
        p["typekit_id"] = r["website"]["typekitId"]
    except KeyError as e:
        p["typekit_id"] = None
    try:
        p["address_title"] = r["website"]["location"][
            "addressTitle"]
    except KeyError as e:
        p["address_title"] = None
    try:
        p["address_line_1"] = r["website"]["location"][
            "addressLine1"]
    except KeyError as e:
        p["address_line_1"] = None
    try:
        p["address_line_2"] = r["website"]["location"][
            "addressLine2"]
    except KeyError as e:
        p["address_line_2"] = None
    try:
        p["latitude"] = r["website"]["location"]["mapLat"]
    except KeyError as e:
        p["latitude"] = None
    try:
        p["longitude"] = r["website"]["location"]["mapLng"]
    except KeyError as e:
        p["latitude"] = None
    # try:
    #     p["payment_settings"] = r["website"][""]

    return p





"""
wordpress_lookup_service.py: A collection of utilities for the wordpress plugin
lookup service.
"""
__author__ = "Jason Wolosonovich <jason@avaland.io>"
__license__ = "BSD 3 Clause"
import re

import requests
from bs4 import BeautifulSoup

from .crawler_service import HEADERS


def get_wp_plugin_info_online(plugin=None):
    """Takes a plugin name and returns the description, name and
    tagshttps://www.silvabokis.com/squarespace-tips/how-to-find-out-which-squarespace-template-a-site-is-using for the plugin."""
    print("get_wp_plugin_info_online()")
    plugin_url = "http://plugins.svn.wordpress.org/{}/trunk/" \
        .format(plugin)
    try:
        r = requests.get(url=plugin_url,
                         headers=HEADERS)
    except requests.exceptions.HTTPError as e:
        print(e)

    if r.status_code != 200:
        description = None
        tags = list()
        name = None
    else:
        # check for the text file as they're potentially named
        # differently
        soup = BeautifulSoup(r.content,
                             "html.parser")

        # find the link with .txt and follow it
        txt_href = soup.select_one("a[href*=.txt]")

        if txt_href is None:
            return list(), None, None
        else:
            txt_href = txt_href.text

        # now visit that and parse the description
        try:
            readme_url = plugin_url + txt_href
            r = requests.get(url=readme_url,
                             headers=HEADERS)
        except requests.exceptions.HTTPError as e:
            print(e)

        if r.status_code != 200:
            description = None
            tags = list()
            name = None
        else:
            name_re = r"(?m)^===\s*(.+?)\s*="
            tags_re = r"(?m)^Tags:\s*(.+?)\r?$"
            d_re = r"(?ms)^== Description ==\s*(.+?)\s+^="
            try:
                name = re.search(name_re,
                                 r.text).group(1)
                name = name.replace("\r", " ").replace("\n", " ")
            except AttributeError as e:
                name = None
            try:
                tags = re.search(tags_re,
                                 r.text)\
                .group(1)\
                .split(",")
                tags = [tag.replace("\r", "").replace("\n", "")
                        for tag in tags]
            except AttributeError as e:
                tags = list()
            try:
                description = re.search(d_re,
                                        r.text).group(1)

                description = description.replace("\r", " ").replace(
                    "\n", " ")
            except AttributeError as e:
                description = None
    print("tags: {}".format(tags))
    print("description: {}".format(description))
    print("name: {}".format(name))
    return tags, description, name



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from rftk.cloud_functions import services

setup(name="the_refinery_toolkit",
      version=services.__version__,
      description="The Refinery Toolkit (RFTK)",
      author="Jason Wolosonovich",
      author_email="jason@avaland.io",
      url="https://github.com/jmwoloso/rftk",
      packages=["rftk.cloud_functions",
                "rftk.cloud_functions.services"],
      license="BSD 3 Clause",
      long_description="The Refinery Toolkit - "
                       "A Collection of Utilities for use "
                       "with The Refinery on GCP",
      install_requires=[
          "beautifulsoup4==4.6.3",
          "cachetools==2.1.0",
          "certifi==2018.10.15",
          "chardet==3.0.4",
          "clearbit==0.1.7",
          "Click==7.0",
          "decorator==4.3.0",
          "dnspython==1.15.0",
          "gcsfs==0.2.0",
          "google-auth-oauthlib==0.2.0",
          "google-api-core==1.6.0a1",
          "google-api-python-client==1.7.4",
          "google-auth==1.5.1",
          "google-auth-httplib2==0.0.3",
          "google-cloud-bigquery==1.6.0",
          "google-cloud-bigquery-datatransfer==0.1.1",
          "google-cloud-core==0.28.1",
          "google-cloud-language==1.1.0",
          "google-cloud-pubsub==0.38.0",
          "google-cloud-storage==1.13.0",
          "google-resumable-media==0.3.1",
          "googleapis-common-protos==1.5.5",
          "grpc-google-iam-v1==0.11.4",
          "grpcio==1.16.0",
          "httplib2==0.11.3",
          "idna==2.7",
          "oauthlib==2.1.0",
          "protobuf==3.6.1",
          "psutil==5.4.7",
          "pyasn1==0.4.4",
          "pyasn1-modules==0.2.2",
          "pytz==2018.6",
          "requests==2.20.0",
          "requests-oauthlib==1.0.0",
          "rsa==4.0",
          "six==1.11.0",
          "typed-ast==1.1.0",
          "uritemplate==3.0.0",
          "urllib3==1.24",
          "validators==0.12.2"
      ],
      classifiers=[
          "Development Status :: 3 - Alpha",
          # "Programming Language :: Python :: 2",
          # "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
          "Topic :: Software Development :: Libraries :: Python "
          "Modules",
          "Topic :: Scientific/Engineering",
          "Operating System :: Unix",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Microsoft :: Windows",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Other Audience",
          "Intended Audience :: Information Technology"
      ])



"""
classes.py: utility classes used in the the Refinery.

"""
import json

from googleapiclient.discovery_cache.base import Cache as GoogleCache

__author__ = "Jason Wolosonovich <jason@avaland.io>"
__license__ = "BSD 3 clause"


class MetadataMixin(dict):
    """Generic mixin class that just serves as a container and can
    have arbitrary attributes programmatically set during refining
    and enrichment."""
    def __init__(self,*arg,**kw):
        super(MetadataMixin, self).__init__(*arg, **kw)

    def __repr__(self):
        return json.dumps(self,
                          indent=2)


class MemoryCache(GoogleCache):
    """Avoids the warning about the file_cache being unavailable."""
    _CACHE = {}

    def get(self, url):
        return MemoryCache._CACHE.get(url)

    def set(self, url, content):
        MemoryCache._CACHE[url] = content



"""
clearbit_service.py: Utility functions used by the clearbit service.
"""
__author__ = "Jason Wolosonovich <jason@avaland.io>"
__license__ = "BSD 3 clause"
import os
import base64
import time

import requests
import googleapiclient.discovery
import clearbit
import gcsfs

from .classes import MemoryCache, MetadataMixin
from .constants import MAX_RETRIES, SLEEP_LENGTH


def decrypt_with_kms(project_id=None, location_id=None,
                     key_ring_id=None, crypto_key_id=None,
                     ciphertext_string=None):
    """Decrypts data from ciphertext_string stored in GCS."""
    print("decrypt_with_kms()")
    # Creates an API client for the KMS API.
    kms_client = googleapiclient\
        .discovery \
        .build(
        "cloudkms",
        "v1",
        cache=MemoryCache()
    )

    # The resource name of the CryptoKey.
    name = \
        "projects/{}/locations/{}/keyRings/{}/cryptoKeys/{}" \
            .format(
            project_id,
            location_id,
            key_ring_id,
            crypto_key_id
        )

    # Use the KMS API to decrypt the data.
    crypto_keys = \
        kms_client \
            .projects() \
            .locations() \
            .keyRings() \
            .cryptoKeys()

    request = crypto_keys.decrypt(
        name=name,
        body={
            "ciphertext": base64.b64encode(
                ciphertext_string
            ).decode(
                "ascii"
            )
        }
    )
    response = request.execute()
    plaintext = base64.b64decode(
        response["plaintext"]
            .encode(
            "ascii"
        )
    ) \
        .decode() \
        .strip("\n")

    return plaintext


def get_service_configs(service=None, project_name=None):
    """Utility function to set configurations for the service."""
    print("get_service_configs()")
    if service == "clearbit":
        try:
            # set configs from the env vars of the machine
            PROJECT_ID = os.environ["GCP_PROJECT_ID"]
            LOCATION_ID = os.environ["GCP_LOCATION_ID"]
            KEYRING_ID = os.environ["GCP_KEYRING_ID"]
            CRYPTO_KEY_ID = os.environ["GCP_KEY_ID"]
            CIPHERTEXT_BLOB = os.environ["GCP_CIPHERTEXT_BLOB"]
            BUCKET_NAME = os.environ["GCS_BUCKET"]

            # DEPRECATED
            # set clearbit api versions
            # clearbit.Person.set_version("2018-06-06")
            # clearbit.Company.set_version("2017-09-12")
            # clearbit.Reveal.set_version("2018-03-28")
            # # not in use
            # clearbit.Watchlist.set_version("2015-11-13")
            # # not in use
            # clearbit.Prospector.set_version("2016-10-04")

            fs = gcsfs.GCSFileSystem(
                project=project_name,
                access="full_control",
                token="cloud",
                # consistency="md5",
                cache_timeout=None,
                secure_serialize=True,
                check_connection=True
            )

            fs.retries = 7
            fs.connect(method="cloud")

            cipher_string = fs.cat(BUCKET_NAME + "/" + CIPHERTEXT_BLOB)

            # DEPRECATED
            # download the file as a string in-memory
            # st_client = storage.Client()
            # bucket = st_client.get_bucket(BUCKET_NAME)
            # cipher_blob = bucket.blob(CIPHERTEXT_BLOB)
            # cipher_string = cipher_blob.download_as_string()

            # decrypt the kms key stored in gcs and set the key attr
            clearbit.key = decrypt_with_kms(
                project_id=PROJECT_ID,
                location_id=LOCATION_ID,
                key_ring_id=KEYRING_ID,
                crypto_key_id=CRYPTO_KEY_ID,
                ciphertext_string=cipher_string
            )
            print("configs_set: True")
        # catch-all
        except Exception as e:
            print("config_set: False")
            error = {
                "error_message": e,
                "status_code": "Unknown"
            }
            print(error)
    # flags for services
    if service == "crawler":
        # NOT IMPLEMENTED/NEEDED
        pass


# TODO: to rftk.clearbit_service module
def make_clearbit_company_gcs_payload(request=None):
    """Utility function to flatten and parse the fields we need in
    BQ."""
    print("make_clearbit_company_gcs_payload()")
    r = request["company"].copy()
    p = MetadataMixin()

    # add the metadata we injected to the request along the way
    p["refinery_company_id"] = request["refinery_company_id"]
    p["refinery_person_id"] = request["refinery_person_id"]
    p["refined_at"] = request["refined_at"]
    p["refined_date"] = request["refined_date"]
    p["sfdc_lead_id"] = request["sfdc_lead_id"]
    p["sfdc_contact_id"] = request["sfdc_contact_id"]
    p["sfdc_asset_id"] = request["sfdc_asset_id"]
    p["sfdc_acct_id"] = request["sfdc_acct_id"]
    p["sfdc_oppty_id"] = request["sfdc_oppty_id"]
    p["app_name"] = request["app_name"]
    p["url"] = request["url"]
    p["domain"] = request["domain"]
    p["ip_revealed"] = request["ip_revealed"]
    p["fuzzy_match"] = request["fuzzy_match"]
    p["ip_address"] = request["ip_address"]

    # add the company data returned by clearbit
    p["clearbit_company_id"] = r["id"]
    p["company_name"] = r["name"]
    p["legal_name"] = r["legalName"]
    p["company_domain"] = r["domain"]
    p["domain_aliases"] = " >>> ".join(
        r["domainAliases"]
    )

    # category
    p["industry"] = r["category"]["industry"]
    p["industry_group"] = r["category"]["industryGroup"]
    p["naics_code"] = r["category"]["naicsCode"]
    p["sector"] = r["category"]["sector"]
    p["sic_code"] = r["category"]["sicCode"]
    p["sub_industry"] = r["category"]["subIndustry"]

    # p["tags"] = r["tags"]
    # TODO: parse this to remove /r
    if r["description"] is not None:
        p["description"] = r["description"].replace("\r", " ")
    else:
        p["description"] = r["description"]
    p["year_founded"] = r["foundedYear"]
    p["location"] = r["location"]
    p["time_zone"] = r["timeZone"]
    p["utc_offset"] = r["utcOffset"]

    # geo
    p["street_number"] = r["geo"]["streetNumber"]
    p["street_name"] = r["geo"]["streetName"]
    p["sub_premise"] = r["geo"]["subPremise"]
    p["city"] = r["geo"]["city"]
    p["postal_code"] = r["geo"]["postalCode"]
    p["state"] = r["geo"]["state"]
    p["state_code"] = r["geo"]["stateCode"]
    p["country"] = r["geo"]["country"]
    p["country_code"] = r["geo"]["countryCode"]
    p["latitude"] = r["geo"]["lat"]
    p["longitude"] = r["geo"]["lng"]

    p["company_logo"] = r["logo"]

    # fakebook
    p["facebook_handle"] = r["facebook"]["handle"]
    p["facebook_likes"] = r["facebook"]["likes"]

    # linkedin
    p["linkedin_handle"] = r["linkedin"]["handle"]

    # twitter
    # TODO: parse the text from bio to eliminate /r
    p["twitter_handle"] = r["twitter"]["handle"]
    p["twitter_id"] = r["twitter"]["id"]
    if r["twitter"]["bio"] is not None:
        p["twitter_bio"] = r["twitter"]["bio"].replace("\r", " ")
    else:
        p["twitter_bio"] = r["twitter"]["bio"]
    p["twitter_follower_count"] = r["twitter"]["followers"]
    p["twitter_following_count"] = r["twitter"]["following"]
    p["twitter_location"] = r["twitter"]["location"]
    p["twitter_site_url"] = r["twitter"]["site"]
    p["twitter_avatar"] = r["twitter"]["avatar"]

    # crunchbase
    p["crunchbase_handle"] = r["crunchbase"]["handle"]

    p["is_email_provider"] = r["emailProvider"]
    p["company_type"] = r["type"]
    p["ticker_symbol"] = r["ticker"]
    p["tax_ein"] = r["identifiers"]["usEIN"]
    p["company_phone"] = r["phone"]
    p["clearbit_indexed_at"] = r["indexedAt"]

    # metrics
    p["alexa_us_rank"] = r["metrics"]["alexaUsRank"]
    p["alexa_global_rank"] = r["metrics"]["alexaGlobalRank"]
    p["number_of_employees"] = r["metrics"]["employees"]
    p["number_of_employees_range"] = r["metrics"]["employeesRange"]
    p["market_cap"] = r["metrics"]["marketCap"]
    p["total_raised"] = r["metrics"]["raised"]
    p["annual_revenue"] = r["metrics"]["annualRevenue"]
    p["estimated_annual_revenue"] = r["metrics"][
        "estimatedAnnualRevenue"]
    p["fiscal_year_ends"] = r["metrics"]["fiscalYearEnd"]

    # p["tech"] = r["tech"]
    p["parent_domain"] = r["parent"]["domain"]
    return p


def make_clearbit_person_gcs_payload(request=None):
    """Utility function to flatten and parse the fields we need in
    BQ."""
    print("make_clearbit_person_gcs_payload()")
    r = request["person"].copy()
    p = MetadataMixin()

    # add the metadata we injected to the request along the way
    p["refinery_person_id"] = request["refinery_person_id"]
    p["refinery_company_id"] = request["refinery_company_id"]
    p["refined_at"] = request["refined_at"]
    p["refined_date"] = request["refined_date"]
    p["sfdc_lead_id"] = request["sfdc_lead_id"]
    p["sfdc_contact_id"] = request["sfdc_contact_id"]
    p["sfdc_asset_id"] = request["sfdc_asset_id"]
    p["sfdc_acct_id"] = request["sfdc_acct_id"]
    p["sfdc_oppty_id"] = request["sfdc_oppty_id"]
    p["app_name"] = request["app_name"]
    p["url"] = request["url"]
    p["domain"] = request["domain"]
    p["ip_revealed"] = request["ip_revealed"]

    p["clearbit_person_id"] = r["id"]
    p["clearbit_indexed_at"] = r["indexedAt"]
    p["full_name"] = r["name"]["fullName"]
    p["first_name"] = r["name"]["givenName"]
    p["last_name"] = r["name"]["familyName"]
    p["email"] = r["email"]

    # geo
    p["location"] = r["location"]
    p["time_zone"] = r["timeZone"]
    p["utc_offset"] = r["utcOffset"]
    p["city"] = r["geo"]["city"]
    p["state"] = r["geo"]["state"]
    p["state_code"] = r["geo"]["stateCode"]
    p["country"] = r["geo"]["country"]
    p["country_code"] = r["geo"]["countryCode"]
    p["latitude"] = r["geo"]["lat"]
    p["longitude"] = r["geo"]["lng"]

    # social
    if r["bio"] is not None:
        p["bio"] = r["bio"].replace("\r", " ")
    else:
        p["bio"] = r["bio"]
    p["site"] = r["site"]
    p["avatar"] = r["avatar"]

    # employment
    p["employment_domain"] = r["employment"]["domain"]
    p["employment_name"] = r["employment"]["name"]
    p["employment_title"] = r["employment"]["title"]
    p["employment_role"] = r["employment"]["role"]

    # more social
    p["employment_seniority"] = r["employment"]["seniority"]
    p["facebook_handle"] = r["facebook"]["handle"]
    p["github_handle"] = r["github"]["handle"]
    p["github_avatar"] = r["github"]["avatar"]
    p["github_company"] = r["github"]["company"]
    p["github_blog"] = r["github"]["blog"]
    p["github_followers"] = r["github"]["followers"]
    p["github_following"] = r["github"]["following"]
    p["twitter_handle"] = r["twitter"]["handle"]
    p["twitter_id"] = r["twitter"]["id"]
    p["twitter_bio"] = r["twitter"]["bio"]
    p["twitter_followers"] = r["twitter"]["followers"]
    p["twitter_following"] = r["twitter"]["following"]
    p["twitter_location"] = r["twitter"]["location"]
    p["twitter_site"] = r["twitter"]["site"]
    p["twitter_avatar"] = r["twitter"]["avatar"]
    p["linkedin_handle"] = r["linkedin"]["handle"]
    p["googleplus_handle"] = r["googleplus"]["handle"]
    p["gravatar_handle"] = r["gravatar"]["handle"]
    p["gravatar_url_titles"] = list()
    p["gravatar_urls"] = list()
    if r["gravatar"]["urls"] is not None:
        for d in r["gravatar"]["urls"]:
            try:
                p["gravatar_url_titles"].append(d["title"])
            except KeyError as e:
                p["gravatar_url_titles"].append(None)
            try:
                p["gravatar_urls"].append(d["value"])
            except KeyError as e:
                p["gravatar_urls"].append(None)

    if len(p["gravatar_url_titles"]) != 0:
        if None in p["gravatar_url_titles"]:
            p["gravatar_url_titles"] = \
                list(set(p["gravatar_url_titles"]))
            p["gravatar_url_titles"].remove(None)
            if len(p["gravatar_url_titles"]) != 0:
                p["gravatar_url_titles"] = " >>> ".join(
                    p["gravatar_url_titles"]
                )
            else:
                p["gravatar_url_titles"] = None
        else:
            p["gravatar_url_titles"] = " >>> ".join(
                p["gravatar_url_titles"]
            )
    else:
        p["gravatar_url_titles"] = None

    if len(p["gravatar_urls"]) != 0:
        if None in p["gravatar_urls"]:
            p["gravatar_urls"] = list(set(p["gravatar_urls"]))
            p["gravatar_urls"].remove(None)
            if len(p["gravatar_urls"]) != 0:
                p["gravatar_urls"] = " >>> ".join(
                    p["gravatar_urls"]
                )
            else:
                p["gravatar_urls"] = None
        else:
            p["gravatar_urls"] = " >>> ".join(
                p["gravatar_urls"]
            )
    else:
        p["gravatar_urls"] = None

    p["gravatar_avatar"] = r["gravatar"]["avatar"]
    p["gravatar_avatar_types"] = list()
    p["gravatar_avatar_urls"] = list()
    if len(r["gravatar"]["avatars"]) != 0:
        for d in r["gravatar"]["avatars"]:
            try:
                p["gravatar_avatar_types"].append(d["type"])
            except KeyError as e:
                p["gravatar_avatar_types"].append(None)
            try:
                p["gravatar_avatar_urls"].append(d["url"])
            except KeyError as e:
                p["gravatar_avatar_urls"].append(None)

    if len(p["gravatar_avatar_types"]) != 0:
        if None in p["gravatar_avatar_types"]:
            p["gravatar_avatar_types"] = \
                list(set(p["gravatar_avatar_types"]))
            p["gravatar_avatar_types"].remove(None)
            if len(p["gravatar_avatar_types"]) != 0:
                p["gravatar_avatar_types"] = " >>> ".join(
                    p["gravatar_avatar_types"]
                )
            else:
                p["gravatar_avatar_types"] = None
        else:
            p["gravatar_avatar_types"] = " >>> ".join(
                p["gravatar_avatar_types"]
            )
    else:
        p["gravatar_avatar_types"] = None

    if len(p["gravatar_avatar_urls"]) != 0:
        if None in p["gravatar_avatar_urls"]:
            p["gravatar_avatar_urls"] = \
                list(set(p["gravatar_avatar_urls"]))
            p["gravatar_avatar_urls"].remove(None)
            if len(p["gravatar_avatar_urls"]) != 0:
                p["gravatar_avatar_urls"] = " >>> ".join(
                    p["gravatar_avatar_urls"]
                )
            else:
                p["gravatar_avatar_urls"] = None
        else:
            p["gravatar_avatar_urls"] = " >>> ".join(
                p["gravatar_avatar_urls"]
            )
    else:
        p["gravatar_avatar_urls"] = None

    p["fuzzy_match"] = r["fuzzy"]
    p["is_email_provider"] = r["emailProvider"]
    return p


# TODO: can we abstract this to fit all cases (tags, tech, wp)
def make_clearbit_tags_payload(request=None):
    """Utility function that creates a payload for the Wordpress
    bucket."""
    print("make_clearbit_tags_payload()")
    print("request: {}".format(request))
    r = request.copy()
    p = MetadataMixin()

    p["refinery_company_id"] = r["refinery_company_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["domain"] = r["domain"]
    p["url"] = r["url"]
    p["ip_revealed"] = r["ip_revealed"]
    p["fuzzy_match"] = r["fuzzy_match"]
    p["tags"] = r["company"]["tags"]
    return p


# TODO: can we abstract this to fit all cases (tags, tech, wp)
def make_clearbit_tech_payload(request=None):
    """Utility function that creates a payload for the Wordpress
    bucket."""
    print("make_clearbit_tech_payload()")
    print("request: {}".format(request))
    r = request.copy()
    p = MetadataMixin()

    p["refinery_company_id"] = r["refinery_company_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["domain"] = r["domain"]
    p["url"] = r["url"]
    p["ip_revealed"] = r["ip_revealed"]
    p["fuzzy_match"] = r["fuzzy_match"]
    p["tech"] = r["company"]["tech"]
    return p


# TODO: can we abstract this to fit all cases (tags, tech, wp)
def make_clearbit_emails_payload(request=None):
    """Utility function that creates a payload for the Wordpress
    bucket."""
    print("make_clearbit_emails_payload()")
    print("request: {}".format(request))
    r = request.copy()
    p = MetadataMixin()

    p["refinery_company_id"] = r["refinery_company_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["domain"] = r["domain"]
    p["url"] = r["url"]
    p["ip_revealed"] = r["ip_revealed"]
    p["fuzzy_match"] = r["fuzzy_match"]
    p["emails"] = r["company"]["site"]["emailAddresses"]
    return p


# TODO: can we abstract this to fit all cases (tags, tech, wp)
def make_clearbit_phones_payload(request=None):
    """Utility function that creates a payload for the Wordpress
    bucket."""
    print("make_clearbit_phones_payload()")
    print("request: {}".format(request))
    r = request.copy()
    p = MetadataMixin()

    p["refinery_company_id"] = r["refinery_company_id"]
    p["refined_at"] = r["refined_at"]
    p["refined_date"] = r["refined_date"]
    p["domain"] = r["domain"]
    p["url"] = r["url"]
    p["ip_revealed"] = r["ip_revealed"]
    p["fuzzy_match"] = r["fuzzy_match"]
    p["phones"] = r["company"]["site"]["phoneNumbers"]
    return p


def check_clearbit_rate_limit(response=None):
    """Determines whether we've exceeded the rate limit."""
    print("check_clearbit_rate_limit()")
    print("headers: {}".format(response.headers))
    rate_limit = response.headers["x-ratelimit-limit"]
    remaining = response.headers["x-ratelimit-remaining"]
    reset_time = response.headers["x-ratelimit-reset"]

    print("rate_limit: {}".format(rate_limit))
    print("requests_remaining: {}".format(remaining))
    print("reset_time: {}".format(reset_time))

    rate_limit_exceeded = False

    if remaining == "0":
        rate_limit_exceeded = True
        print("rate limit exceeded: {}".format(rate_limit_exceeded))
    results = {
        "rate_limit_exceeded": rate_limit_exceeded,
        "reset_time": reset_time,
    }
    print("results: {}".format(results))

    return results


def clearbit_routines(request=None, streaming=False):
    """Enrichment routines that wrap the Clearbit API."""
    print("clearbit_routines()")
    # TODO: add functionality to enable revealing and then enriching
    #  make the appropriate endpoint request
    # TODO: make sure to return the expected payload here
    CONTINUE = True
    n = 0

    while CONTINUE and n < MAX_RETRIES:
        try:
            if request["clearbit_reveal"] is True:
                print("Reveal API")
                # ip_address = request["ip_address"]
                print("ip_address: {}".format(request["ip_address"]))

                # TODO: find out if there is a streaming endpoint for
                #  Reveal
                if streaming is True:
                    resp = requests.get(
                        url=
                        "https://reveal.clearbit.com/v1/companies/find",
                        params={"ip": request["ip_address"]},
                        auth=(clearbit.key, ""),
                        headers={
                            "API-Version": "2018-03-28"
                        }
                    )
                else:
                    resp = requests.get(
                        url=
                        "https://reveal.clearbit.com/v1/companies/find",
                        params={"ip": request["ip_address"]},
                        auth=(clearbit.key, ""),
                        headers={
                            "API-Version": "2018-03-28"
                        }
                    )
                n += 1

                # DEPRECATED
                # resp = clearbit.Reveal.find(ip=ip_address,
                #                             streaming=streaming)


                # TODO: figure out how to implement this
                # rate limit check (we'll pause with this request to
                # control the flow of future requests
                results = check_clearbit_rate_limit(
                    response=resp
                )
                if results["rate_limit_exceeded"] is True:
                    reset_time = int(results["reset_time"])
                    current_time = time.time()
                    sleep_time = reset_time - current_time
                    if sleep_time > 0:
                        time.sleep(sleep_time + 1)
                        continue
                    else:
                        pass
                else:
                    print("still within rate limit.")

                print("request sent")
                print("status: {}".format(resp.status_code))
                # 202 (async lookup; try again momentarily)
                if "pending" in resp.json().keys():
                    print("request is pending.")
                    time.sleep(SLEEP_LENGTH)
                    continue

                # 404
                if resp is None:
                    print("resp: {}".format(resp))
                    err = {
                        "error_message": "Person/Company Not Found",
                        "status_code": "404"
                    }
                    print(err)

                    # we still need to return the expected payload
                    resp = {
                        "error": "Person/Company Not Found",
                        "status": "404",
                        "company": None,
                        "person": None
                    }
                break
            elif request["clearbit_enrich"] is True:
                # TODO: NOTE: big "gotcha" here as the email address
                #  takes precedent over the domain if provided,
                #  so using something like
                #  jason.wolosonovich@infusionsoft.com with a domain
                #  of 'cxmybiz.com' will return person info on jason
                #  as well as company info on infusionsoft (instead
                #  of cxmybiz.com)
                print("Enrichment API")
                print("email_address: {}".format(request["email"]))
                params = {
                    "email": request["email"],
                    "domain": request["domain"]
                }

                # determine from the payload which endpoint to hit
                enrichment_api = "person" if request["email"] is not \
                                             None else "company"
                print("enrichment api: {}".format(enrichment_api))
                print("sending request")
                # TODO: implement manual streaming
                if streaming is True:
                    if enrichment_api == "person":
                        resp = requests.get(
                            url="https://person-stream.clearbit.com/v2"
                                "/combined/find",
                            params={**params},
                            auth=(clearbit.key, ""),
                            headers={
                                "API-Version": "2018-11-19"
                            }
                        )

                    elif enrichment_api == "company":
                        resp = requests.get(
                            url="https://company-stream.clearbit.com/v2"
                                "/companies/find",
                            params={**params},
                            auth=(clearbit.key, ""),
                            headers={
                                "API-Version": "2017-09-12"
                            }
                        )
                else:
                    if enrichment_api == "person":
                        resp = requests.get(
                            url="https://person.clearbit.com/v2/combined/find",
                            params={**params},
                            auth=(clearbit.key, ""),
                            headers={
                                "API-Version": "2018-11-19"
                            }
                        )
                    elif enrichment_api == "company":
                        resp = requests.get(
                            url="https://company.clearbit.com/v2/companies/find",
                            params={**params},
                            auth=(clearbit.key, ""),
                            headers={
                                "API-Version": "2017-09-12"
                            }
                        )
                n += 1

                # TODO: figure out how to do this
                # rate limit check (we'll pause with this request to
                # control the flow of future requests
                results = check_clearbit_rate_limit(
                    response=resp
                )

                if results["rate_limit_exceeded"] is True:
                    reset_time = int(results["reset_time"])
                    current_time = time.time()
                    sleep_time = reset_time - current_time
                    if sleep_time > 0:
                        time.sleep(sleep_time + 1)
                        continue
                    else:
                        pass
                else:
                    print("still within rate limit.")

                print("request sent")
                print("status: {}".format(resp.status_code))

                # 202 (async lookup; try again momentarily)
                if resp.status_code == 201 or resp.status_code == 202:
                    print("request is pending.")
                    time.sleep(SLEEP_LENGTH)
                    continue
                # if "pending" in resp.keys():
                #     print("request is pending.")
                #     time.sleep(SLEEP_LENGTH)
                #     continue

                # 404
                if resp is None:
                    print("resp: {}".format(resp))
                    err = {
                        "error_message": "Person/Company Not Found",
                        "status_code": "404"
                    }
                    print(err)

                    # we still need to return the expected payload
                    resp = {
                        "error": "Person/Company Not Found",
                        "status": "404",
                        "company": None,
                        "person": None
                    }
                resp.raise_for_status()
                break
            elif request["clearbit_reveal"] is False and request[
                "clearbit_enrich"] is False:
                print("either `clearbit_reveal` or `clearbit_enrich` "
                      "must be True")
                try:
                    r = requests.Response()
                    r.status_code = 422
                    r.raise_for_status()
                except requests.exceptions.HTTPError as e:
                    print("Unprocessable Entity, 422")
                    raise(requests.exceptions.HTTPError(e))
        except requests.exceptions.HTTPError as e:
            print(e)
            status = str(e.response.status_code)
            if status == "400":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Bad Request",
                        "status": "400",
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("400 error.")
                    break
                continue
            if status == "401":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Invalid API Key",
                        "status": "401",
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("401 error.")
                    break
                continue
            if status == "402":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Quota Exceeded",
                        "status": "402",
                        "company": None,
                        "person": None,
                        "quota_info": resp.headers
                    }
                    print(resp)
                    print("402 error.")
                    break
                continue
            if status == "404":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Person/Company Not Found",
                        "status": "404",
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("404 error.")
                    break
                continue

            if status == "422":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Unprocessable Entity",
                        "status": "422",
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("422 error.")
                    break
                continue
            elif status == "429":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Client Error: Too Many Requests",
                        "status": "429",
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("429 error.")
                    break
                continue
            elif status[0] == "5":
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "Internal Server Error",
                        "status": "5xx",
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("500 error.")
                    break
                continue
            else:
                if n == MAX_RETRIES - 1:
                    # we still need to return the expected payload
                    resp = {
                        "error": "{}".format(status),
                        "status": "{}".format(status),
                        "company": None,
                        "person": None
                    }
                    print(resp)
                    print("MAX_RETRIES EXCEEDED")
                    break
                continue
        except UnboundLocalError as e:
            print(e)
            if n == MAX_RETRIES - 1:
                resp = {
                    "error": "Person/Company Not Found",
                    "status": "404",
                    "company": None,
                    "person": None
                }
                print(resp)
                print("Unbound Local Error.")
                break
            continue
        # catch-all for python exceptions
        except Exception as e:
            print(e)
            if n == MAX_RETRIES - 1:
                resp = {
                    "error": "Person/Company Not Found",
                    "status": "404",
                    "company": None,
                    "person": None
                }
                print(resp)
                print("Error: {}".format(e))
                break
            continue          
    
        # TODO: this is really a 404 error but the clearbit
        # library makes it unclear that this is the case
        # ref: https://github.com/clearbit/clearbit-python/blob
        # /master/clearbit/resource.py
    try:
        if resp.status_code == 200:
            print("request succeeded")
            resp = resp.json()
        if resp.status_code == 202:
            print("request was pending for too long.")
            resp = resp.json()
            # we'll flag this as an error
            resp["error"] = True
    except AttributeError as e:
        pass
    print("returning response.")
    return resp
